
#include "X386.h"
#include "vga.h"

#ifndef SINGLEWINDOW
	error - tvga8900 requires SINGLEWINDOW to be defined
#endif

typedef struct {
	vgaHWRec std;          /* std IBM VGA register */
	unsigned char mode;
	unsigned char New1;
	unsigned char New2;
	unsigned char Old1;
	unsigned char Old2;
	unsigned char Seq12;
	unsigned char CRTCModuleTest;
} vgaTVGA8900Rec, *vgaTVGA8900Ptr;

static void TVGA8900Init();
static void *TVGA8900Save();
static void TVGA8900Restore();
static void TVGA8900Adjust();
extern void TVGA8900SetReadWrite();


vgaVideoChipRec TVGA8900 = {
  TVGA8900Init,
  TVGA8900Save,
  TVGA8900Restore,
  TVGA8900Adjust,
  TVGA8900SetReadWrite,
  TVGA8900SetReadWrite,
  TVGA8900SetReadWrite,
  0x10000,
  0x10000,
  16,
  0xffff,
  0x00000, 0x10000,
  0x00000, 0x10000,
};

/*
 *-----------------------------------------------------------------------
 * TVGA8900Restore --
 *      restore a video mode
 *
 * Results:
 *      nope.
 *
 * Side Effects: 
 *      the display enters a new graphics mode. 
 *
 *-----------------------------------------------------------------------
 */

static void
TVGA8900Restore(restore)
     vgaTVGA8900Ptr restore;
{
  TRACE( ("TVGA8900Restore(restore=0x%x)\n",restore) );

	outb(0x3c4,11);
	inb(0x3c5);
	outw(0x3c4,0x820e);
	outw(0x3c4,restore->Seq12<<8 | 12);
	outw(0x3c4,0x20e);
	
	outw(0x3c4,11);

	outw(0x3c4,restore->Old2<<8 | 13);
	outw(0x3c4,(restore->Old1^2)<<8 | 14);

	outb(0x3c4,11);
	inb(0x3c5);
	outw(0x3c4,restore->New2<<8 | 13);
	outw(0x3c4,(restore->New1^2)<<8 | 14);

	outw(vgaIOBase+4,restore->CRTCModuleTest << 8 | 30);

	if (!restore->mode)
		outw(0x3c4,11);

  vgaHWRestore(restore);

	outw(0x3C4, 0x0300); /* now reenable the timing sequencer */
	
}

/*
 *-----------------------------------------------------------------------
 * TVGA8900Save --
 *      save the current video mode
 *
 * Results:
 *      pointer to the current mode record.
 *
 * Side Effects: 
 *      None.
 *-----------------------------------------------------------------------
 */
static void *
TVGA8900Save(save)
     vgaTVGA8900Ptr save;
{
	int i,j;

  TRACE( ("TVGA8900Save(save=0x%x)\n",save) );

  save = (vgaTVGA8900Ptr)vgaHWSave(save, sizeof(vgaTVGA8900Rec));

#define new ((vgaTVGA8900Ptr)vgaNewVideoState)

	outb(0x3c4,14);
	i = inb(0x3c5);
	outb(0x3c4,13);
	j = inb(0x3c5);

	outw(0x3c4,11);
	outb(0x3c4,14); new->Old1 = inb(0x3c5);
	outb(0x3c4,13); new->Old2 = inb(0x3c5);

	outb(0x3c4,11);
	inb(0x3c5);
	outb(0x3c4,14); new->New1 = inb(0x3c5);
	outb(0x3c4,13); new->New2 = inb(0x3c5);
	outb(0x3c4,12); new->Seq12 = inb(0x3c5);

/* if register1 or 2 changes when we switch to old mode, it was probably
in new mode, unless the new and old registers are identical, in which case
we assume it was new mode */
	new->mode = new->Old1!=i || new->Old2!=j || (new->New1==i && new->New2==j);

	outb(vgaIOBase+4,30); new->CRTCModuleTest = inb(vgaIOBase+5);

  return ((void *) save);
}

/*
 *-----------------------------------------------------------------------
 * TVGA8900Init --
 *      Handle the initialization, etc. of a screen.
 *
 * Results:
 *      pointer to the new mode record.
 *
 * Side Effects: 
 *
 *-----------------------------------------------------------------------
 */
static void
TVGA8900Init(mode)
vgaModePtr mode;
{
	int newVTotal,newVSyncStart,newVSyncEnd,newVDisplay;

	TRACE( ("TVGA8900Init(mode=0x%x)\n",mode) );

#define new ((vgaTVGA8900Ptr)vgaNewVideoState)
  
	vgaHWInit(mode,sizeof(vgaTVGA8900Rec));

	newVTotal = mode->VTotal;
	newVSyncStart = mode->VSyncStart;
	newVSyncEnd = mode->VSyncEnd;
	newVDisplay = mode->VDisplay;

	if (mode->Flags & V_INTERLACE) {
		newVTotal >>= 1;
		newVSyncStart >>= 1;
		newVSyncEnd >>= 1;
		newVDisplay >>= 1;
	}

  new->std.CRTC[6]  = (newVTotal - 2) & 0xFF;
  new->std.CRTC[7]  = (((newVTotal -2) & 0x100) >> 8 )
    | (((newVDisplay -1) & 0x100) >> 7 )
      | ((newVSyncStart & 0x100) >> 6 )
	| (((newVSyncStart) & 0x100) >> 5 )
	  | 0x10
	    | (((newVTotal -2) & 0x200)   >> 4 )
	      | (((newVDisplay -1) & 0x200) >> 3 )
		| ((newVSyncStart & 0x200) >> 2 );
  new->std.CRTC[9]  = ((newVSyncStart & 0x200) >>4) | 0x40;
  new->std.CRTC[16] = newVSyncStart & 0xFF;
  new->std.CRTC[17] = (newVSyncEnd & 0x0F) | 0x20;
  new->std.CRTC[18] = (newVDisplay -1) & 0xFF;
  new->std.CRTC[21] = newVSyncStart & 0xFF; 
  new->std.CRTC[22] = (newVSyncStart +1) & 0xFF;


	new->std.CRTC[19] = X386Screen.VirtualX>>(mode->Flags&V_INTERLACE?3:4);
	new->std.CRTC[20] = 0x40;

	new->New1 = 0;
	new->New2 = new->std.NoClock >> 2;

	new->Old1 = 0xa8;
	new->Old2 = 0x10;

	new->mode = 1;
	new->Seq12 = 0xec;
	new->CRTCModuleTest = mode->Flags&V_INTERLACE ? 0x84 : 0x80;
}
	

/*
 *-----------------------------------------------------------------------
 * TVGA8900Adjust --
 *      adjust the current video frame to display the mousecursor
 *
 * Results:
 *      nope.
 *
 * Side Effects: 
 *      the display scrolls
 *-----------------------------------------------------------------------
 */

static void 
TVGA8900Adjust()
{
	int base;
	int y;

	TRACE( ("TVGA8900Adjust()\n") );
	base = X386Screen.FrameY0*X386Screen.VirtualX+X386Screen.FrameX0 >> 3;
  	outw(vgaIOBase+4,(base&255)<<8 | 13);
  	outw(vgaIOBase+4,(base&0xFF00) | 12);
	outb(vgaIOBase+4,30);
	outb(vgaIOBase+5,(inb(vgaIOBase+5)&~32) | (base>65535 ? 32 : 0));
}
