/*
 * Copyright 1990,91 by Thomas Roell, Dinkelscherben, Germany.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Thomas Roell not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Thomas Roell makes no representations
 * about the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * THOMAS ROELL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THOMAS ROELL BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Thomas Roell, roell@informatik.tu-muenchen.de
 *
 * $Header: /proj/X11/mit/server/ddx/at386/vga/RCS/vgaBitBlt.c,v 1.5 91/02/10 16:44:40 root Exp $
 */


#include        "X386.h"
#include        "vga.h"
#include	"tfb.h"


extern void fastBitBltCopy();

#ifdef SINGLEWINDOW

#define BUFSIZE		4096

void
vgaBitBlt(pdstBase, psrcBase, widthSrc, widthDst, x, y,
	    x1, y1, w, h, xdir, ydir, alu, planemask)
     unchar *psrcBase, *pdstBase;  /* start of src and dst bitmaps */
     int    widthSrc, widthDst;
     int    x, y, x1, y1, w, h;
     int    xdir, ydir;
     int    alu;
     ulong  planemask;
{
	unsigned char buf[BUFSIZE];
	int m,n;

	if (alu == GXcopy && (planemask & 0xFF) == 0xFF) {
		if (w == 0)	/* avoid /0 */
			return;
		if ((m=BUFSIZE/w) <= 0)
			return;	/* what else can i do? */
		if (ydir == 1) {
			while (h) {
				n = m>h ? h : m;
				vgaImageRead(buf,psrcBase,widthSrc,w,x,y,0,0,w,n,xdir,ydir,alu,planemask);
				vgaImageWrite(pdstBase,buf,w,widthDst,0,0,x1,y1,w,n,xdir,ydir,alu,planemask);
				y += n;
				y1 += n;
				h -= n;
			}
		}
		else {
			y += h;
			y1 += h;
			while (h) {
				n = m>h ? h : m;
				y -= n;
				y1 -= n;
				h -= n;
				vgaImageRead(buf+BUFSIZE-w,psrcBase,widthSrc,w,x,y,0,0,w,n,xdir,ydir,alu,planemask);
				vgaImageWrite(pdstBase,buf+BUFSIZE-w,w,widthDst,0,0,x1,y1,w,n,xdir,ydir,alu,planemask);
			}
		}
	}
}

#else


void
vgaBitBlt(pdstBase, psrcBase, widthSrc, widthDst, x, y,
	    x1, y1, w, h, xdir, ydir, alu, planemask)
     unchar *psrcBase, *pdstBase;  /* start of src and dst bitmaps */
     int    widthSrc, widthDst;
     int    x, y, x1, y1, w, h;
     int    xdir, ydir;
     int    alu;
     ulong  planemask;

{
  unchar *psrc, *pdst;
  int hcount, vcount, count, srcPitch, dstPitch, tmp;

  if (alu == GXcopy && (planemask & 0xFF) == 0xFF) {
    if (xdir == 1) { /* left to right */
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x;
	  pdst = pdstBase+(y1*widthDst)+x1;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1;
	}
      srcPitch = widthSrc - w;
      dstPitch = widthDst - w;
      
    } else  {/* right to left */
      
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x+w;
	  pdst = pdstBase+(y1 * widthDst)+x1+w;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x+w;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1+w;
	}
      srcPitch = widthSrc + w;
      dstPitch = widthDst + w;
    }
    
    psrc = (unchar *)vgaSetRead(psrc);
    pdst = (unchar *)vgaSetWrite(pdst);
    
    vcount = 0;
    while (h || vcount) {
      /* 
       * compute here the maximal linecount(hcount) before a
       * segment switch has to be made
       */
      if (!vcount) {
	if (ydir == 1) {
	  tmp = ((unchar *)vgaReadTop - psrc) / widthSrc;
	  if ((xdir == -1) &&
	      ((psrc - (unchar *)vgaReadBottom) <= w)) tmp = 0;
	} else {
	  tmp = (psrc - (unchar *)vgaReadBottom) / (-widthSrc);
	  if ((xdir == 1) &&
	      (((unchar *)vgaReadTop - psrc) <= w)) tmp = 0;
	}
	hcount=min( h, tmp );
	if(!hcount) vcount = w;
	
	if (ydir == 1) {
	  tmp = ((unchar *)vgaWriteTop - pdst) / widthDst;
	  if ((xdir == -1) &&
	      ((pdst - (unchar *)vgaWriteBottom) <= w)) tmp = 0;
	} else {
	  tmp = (pdst - (unchar *)vgaWriteBottom) / (-widthDst);
	  if ((xdir == 1) &&
	      (((unchar *)vgaWriteTop - pdst) <= w)) tmp = 0;
	}
	hcount=min( hcount, tmp );
	if(!hcount) vcount = w;
      }

      /*
       * if we now have to make a segment switch within a line,
       * divide this line into small parts.
       */
      if (vcount) {
	count = vcount;
	
	/* check which segment to switch first */
	if (xdir == 1)  {
	  if (tmp = (unchar *)vgaReadTop - psrc)
	    count = min(count, tmp);
	  else
	    psrc = (unchar *)vgaReadNext(psrc);
	} else {
	  if (tmp = psrc - (unchar *)vgaReadBottom)
	    count = min(count, tmp);
	  else
	    psrc = (unchar *)vgaReadPrev(psrc);
	}
	
	if (xdir == 1) {
	  if (tmp = (unchar *)vgaWriteTop - pdst)
	    count = min(count, tmp);
	  else
	    pdst = (unchar *)vgaWriteNext(pdst);
	} else {
	  if (tmp = pdst - (unchar *)vgaWriteBottom)
	    count = min(count, tmp);
	  else
	    pdst = (unchar *)vgaWritePrev(pdst);
	}
	
	fastBitBltCopy(xdir,psrc, pdst, 1, count, 0, 0);
	
	if (!(vcount -= count)) {
	  h--;  /* partial line finish */
	  if (xdir == 1) {
	    psrc += (srcPitch + count);
	    pdst += (dstPitch + count);
	  } else {
	    psrc += (srcPitch - count);
	    pdst += (dstPitch - count);
	  }

	  if ( psrc >= (unchar *)vgaReadTop )
	    psrc = (unchar *)vgaReadNext(psrc);
	  if ( psrc < (unchar *)vgaReadBottom )
	    psrc = (unchar *)vgaReadPrev(psrc);

	  if ( pdst >= (unchar *)vgaWriteTop )
	    pdst = (unchar *)vgaWriteNext(pdst);
	  if ( pdst < (unchar *)vgaWriteBottom )
	    pdst = (unchar *)vgaWritePrev(pdst);

	} else {
	  if (xdir == 1) {
	    psrc += count;
	    pdst += count;
	  } else {
	    psrc -= count;
	    pdst -= count;
	  }
	}
      } else {
	
	h -= hcount;
	
	fastBitBltCopy(xdir, psrc, pdst, hcount, w, srcPitch, dstPitch);
	
	psrc += (hcount*widthSrc);
	pdst += (hcount*widthDst);
      }
    }
  }
}


#endif


void
vgaImageRead(pdstBase, psrcBase, widthSrc, widthDst, x, y,
	    x1, y1, w, h, xdir, ydir, alu, planemask)
     unchar *psrcBase, *pdstBase;  /* start of src and dst bitmaps */
     int    widthSrc, widthDst;
     int    x, y, x1, y1, w, h;
     int    xdir, ydir;
     int    alu;
     ulong  planemask;

{
  register unchar *psrc, *pdst;
  int hcount, vcount, count, srcPitch, dstPitch, tmp;

  if (alu == GXcopy && (planemask & 0xFF) == 0xFF) {
    if (xdir == 1) { /* left to right */
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x;
	  pdst = pdstBase+(y1*widthDst)+x1;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1;
	}
      srcPitch = widthSrc - w;
      dstPitch = widthDst - w;
      
    } else  {/* right to left */
      
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x+w;
	  pdst = pdstBase+(y1 * widthDst)+x1+w;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x+w;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1+w;
	}
      srcPitch = widthSrc + w;
      dstPitch = widthDst + w;
    }
    
    psrc = (unchar *)vgaSetReadWrite(psrc);
    
    vcount = 0;
    while (h || vcount) {
      /* 
       * compute here the maximal linecount(hcount) before a
       * segment switch has to be made
       */
      if (!vcount) {
	if (ydir == 1) {
	  tmp = ((unchar *)vgaWriteTop - psrc) / widthSrc;
	  if ((xdir == -1) &&
	      ((psrc - (unchar *)vgaWriteBottom) <= w)) tmp = 0;
	} else {
	  tmp = (psrc - (unchar *)vgaWriteBottom) / (-widthSrc);
	  if ((xdir == 1) &&
	      (((unchar *)vgaWriteTop - psrc) <= w)) tmp = 0;
	}
	hcount=min( h, tmp );
	if(!hcount) vcount = w;
      }

      /*
       * if we now have to make a segment switch within a line,
       * divide this line into small parts.
       */
      if (vcount) {
	count = vcount;
	
	/* check which segment to switch first */
	if (xdir == 1)  {
	  if (tmp = (unchar *)vgaWriteTop - psrc)
	    count = min(count, tmp);
	  else
	    psrc = (unchar *)vgaReadWriteNext(psrc);
	} else {
	  if (tmp = psrc - (unchar *)vgaWriteBottom)
	    count = min(count, tmp);
	  else
	    psrc = (unchar *)vgaReadWritePrev(psrc);
	}
	
	fastBitBltCopy(xdir,psrc, pdst, 1, count, 0, 0);
	
	if (!(vcount -= count)) {
	  h--;  /* partial line finish */
	  if (xdir == 1) {
	    psrc += (srcPitch + count);
	    pdst += (dstPitch + count);
	  } else {
	    psrc += (srcPitch - count);
	    pdst += (dstPitch - count);
	  }

	  if ( psrc >= (unchar *)vgaWriteTop )
	    psrc = (unchar *)vgaReadWriteNext(psrc);
	  if ( psrc < (unchar *)vgaWriteBottom )
	    psrc = (unchar *)vgaReadWritePrev(psrc);

	} else {
	  if (xdir == 1) {
	    psrc += count;
	    pdst += count;
	  } else {
	    psrc -= count;
	    pdst -= count;
	  }
	}
      } else {
	
	h -= hcount;
	
	fastBitBltCopy(xdir, psrc, pdst, hcount, w, srcPitch, dstPitch);
	
	psrc += (hcount*widthSrc);
	pdst += (hcount*widthDst);
      }
    }
  }
}


void
vgaImageWrite(pdstBase, psrcBase, widthSrc, widthDst, x, y,
	    x1, y1, w, h, xdir, ydir, alu, planemask)
     unchar *psrcBase, *pdstBase;  /* start of src and dst bitmaps */
     int    widthSrc, widthDst;
     int    x, y, x1, y1, w, h;
     int    xdir, ydir;
     int    alu;
     ulong  planemask;

{
  register unchar *psrc, *pdst;
  int hcount, vcount, count, srcPitch, dstPitch, tmp;

  if (alu == GXcopy && (planemask & 0xFF) == 0xFF) {
    if (xdir == 1) { /* left to right */
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x;
	  pdst = pdstBase+(y1*widthDst)+x1;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1;
	}
      srcPitch = widthSrc - w;
      dstPitch = widthDst - w;
      
    } else  {/* right to left */
      
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x+w;
	  pdst = pdstBase+(y1 * widthDst)+x1+w;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x+w;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1+w;
	}
      srcPitch = widthSrc + w;
      dstPitch = widthDst + w;
    }
    
    pdst = (unchar *)vgaSetReadWrite(pdst);
    
    vcount = 0;
    while (h || vcount) {
      /* 
       * compute here the maximal linecount(hcount) before a
       * segment switch has to be made
       */
      if (!vcount) {
	if (ydir == 1) {
	  tmp = ((unchar *)vgaWriteTop - pdst) / widthDst;
	  if ((xdir == -1) &&
	      ((pdst - (unchar *)vgaWriteBottom) <= w)) tmp = 0;
	} else {
	  tmp = (pdst - (unchar *)vgaWriteBottom) / (-widthDst);
	  if ((xdir == 1) &&
	      (((unchar *)vgaWriteTop - pdst) <= w)) tmp = 0;
	}
	hcount=min( h, tmp );
	if(!hcount) vcount = w;
      }

      /*
       * if we now have to make a segment switch within a line,
       * divide this line into small parts.
       */
      if (vcount) {
	count = vcount;
	
	/* check which segment to switch first */
	if (xdir == 1) {
	  if (tmp = (unchar *)vgaWriteTop - pdst)
	    count = min(count, tmp);
	  else
	    pdst = (unchar *)vgaReadWriteNext(pdst);
	} else {
	  if (tmp = pdst - (unchar *)vgaWriteBottom)
	    count = min(count, tmp);
	  else
	    pdst = (unchar *)vgaReadWritePrev(pdst);
	}
	
	fastBitBltCopy(xdir,psrc, pdst, 1, count, 0, 0);
	
	if (!(vcount -= count)) {
	  h--;  /* partial line finish */
	  if (xdir == 1) {
	    psrc += (srcPitch + count);
	    pdst += (dstPitch + count);
	  } else {
	    psrc += (srcPitch - count);
	    pdst += (dstPitch - count);
	  }

	  if ( (void *)pdst >= vgaWriteTop )
	    pdst = (unchar *)vgaReadWriteNext(pdst);
	  if ( (void *)pdst < vgaWriteBottom )
	    pdst = (unchar *)vgaReadWritePrev(pdst);

	} else {
	  if (xdir == 1) {
	    psrc += count;
	    pdst += count;
	  } else {
	    psrc -= count;
	    pdst -= count;
	  }
	}
      } else {
	
	h -= hcount;
	
	fastBitBltCopy(xdir, psrc, pdst, hcount, w, srcPitch, dstPitch);
	
	psrc += (hcount*widthSrc);
	pdst += (hcount*widthDst);
      }
    }
  }
}


void
vgaPixBitBlt(pdstBase, psrcBase, widthSrc, widthDst, x, y,
	    x1, y1, w, h, xdir, ydir, alu, planemask)
     unchar *psrcBase, *pdstBase;  /* start of src and dst bitmaps */
     int    widthSrc, widthDst;
     int    x, y, x1, y1, w, h;
     int    xdir, ydir;
     int    alu;
     ulong  planemask;

{
  register unchar *psrc, *pdst;
  int hcount, vcount, count, srcPitch, dstPitch, tmp;

  if (alu == GXcopy && (planemask & 0xFF) == 0xFF) {
    if (xdir == 1) { /* left to right */
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x;
	  pdst = pdstBase+(y1*widthDst)+x1;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1;
	}
      srcPitch = widthSrc - w;
      dstPitch = widthDst - w;
      
    } else  {/* right to left */
      
      if (ydir == 1) /* top to bottom */
	{
	  psrc = psrcBase+(y*widthSrc)+x+w;
	  pdst = pdstBase+(y1 * widthDst)+x1+w;
	}
      else /* bottom to top */
	{
	  psrc = psrcBase-((y+h-1)*widthSrc)+x+w;
	  pdst = pdstBase-((y1+h-1)*widthDst)+x1+w;
	}
      srcPitch = widthSrc + w;
      dstPitch = widthDst + w;
    }
    
    fastBitBltCopy(xdir, psrc, pdst, h, w, srcPitch, dstPitch);
	
  }
}
