/*
 * Copyright 1990,91 by Thomas Roell, Dinkelscherben, Germany.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Thomas Roell not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Thomas Roell makes no representations
 * about the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * THOMAS ROELL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THOMAS ROELL BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Thomas Roell, roell@informatik.tu-muenchen.de
 *
 * $Header: /proj/X11/mit/server/ddx/at386/vga/RCS/vgaConfig.c,v 1.7 91/02/10 16:44:43 root Exp $
 */

#define X386LineNo 0

#include "X386.h"
#include "vga.h"

#define INTERLACE 3
#define PHSYNC    4
#define NHSYNC    5
#define PVSYNC    6
#define NVSYNC    7
#define FASTRAM   8
#define LEGEND    9
#define SYNCCLUT  10
#define CET3000   100
#define CET4000   101
#define CWDI      102
#define CGVGA     103
#define CATI      104
#define CC452     105
#define CC453     106
#define CTVGA8900     107

static SymTabRec SymTab[] = {
  { INTERLACE, "interlace"},
  { PHSYNC,    "+hsync"},
  { NHSYNC,    "-hsync"},
  { PVSYNC,    "+vsync"},
  { NVSYNC,    "-vsync"},
  { FASTRAM,   "fastram"},
  { LEGEND,    "legend"},
  { CET3000,   "et3000"},
  { CET4000,   "et4000"},
  { CWDI,      "wdi"},
  { CGVGA,     "gvga"},
  { CATI,      "ati"},
  { CC452,     "c452"},
  { CC453,     "c453"},
  { CTVGA8900,     "tvga8900"},
  { -1,        ""},
};

/*
 *-----------------------------------------------------------------------
 * vgaConfig --
 *      Read the rest of the config-file, and initialize the structs
 *      for the vga board according;
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The VGA is initialized.
 *
 *-----------------------------------------------------------------------
 */

void 
vgaConfig()
{
  char *DefaultMode;
  int token,i;
  vgaVideoChipPtr VideoChip;

  vgaChipSetFlags = 0;

  /*
   * get the Chipset we have to deal with.
   */
  switch(X386Lex(SymTab)) {
    case CET3000:   VideoChip = &ET3000; break;
    case CET4000:   VideoChip = &ET4000; break;
    case CWDI:      VideoChip = &WDI; break;
    case CGVGA:     VideoChip = &GVGA;   break;
    case CTVGA8900:     VideoChip = &TVGA8900;   break;
    case CATI:      VideoChip = &ATI;   break;
    case CC452:     VideoChip = &C452;   break;
    case CC453:     VideoChip = &C453;   break;
    default: X386ConfigError("You must sepcify a ChipSet");
  }

  vgaInitFunc = VideoChip->ChipInit;
  vgaSaveFunc = VideoChip->ChipSave;
  vgaRestoreFunc = VideoChip->ChipRestore;
  vgaAdjustFunc = VideoChip->ChipAdjust;
  vgaSetReadFunc = VideoChip->ChipSetRead;
  vgaSetWriteFunc = VideoChip->ChipSetWrite;
  vgaSetReadWriteFunc = VideoChip->ChipSetReadWrite;
  vgaMapSize = VideoChip->ChipMapSize;
  vgaSegmentSize = VideoChip->ChipSegmentSize;
  vgaSegmentShift = VideoChip->ChipSegmentShift;
  vgaSegmentMask = VideoChip->ChipSegmentMask;
  vgaReadBottom = (pointer)VideoChip->ChipReadBottom;
  vgaReadTop = (pointer)VideoChip->ChipReadTop;
  vgaWriteBottom = (pointer)VideoChip->ChipWriteBottom;
  vgaWriteTop = (pointer)VideoChip->ChipWriteTop;

  if (X386Lex(NULL) == NUMBER) 
    vgaVideoRam = X386Val.num;
  else
    X386ConfigError("Bad value for video ram");

  if (X386Lex(NULL) == NUMBER) 
    X386Screen.VirtualX = X386Val.num;
  else
    X386ConfigError("Bad value for virtual x-resolution");

  if (X386Lex(NULL) == NUMBER) 
    X386Screen.VirtualY = X386Val.num;
  else
    X386ConfigError("Bad value for virtual y-resolution");

  /*
   * sanity-check here
   */
  if (X386Screen.VirtualX * X386Screen.VirtualY > vgaVideoRam*1024)
    X386ConfigError("Virtual resolution to big for video ram");

  /*
   * get the clock definitions now
   */
  while (((token=X386Lex(NULL)) == NUMBER) && (vgaVideoClocks < NOCLOCKS))
    {
      vgaVideoClock[vgaVideoClocks++] = X386Val.num;
    }

  if (!vgaVideoClocks) 
    X386ConfigError("You must specify at least on Dot-Clock");

  if (token == STRING)
    DefaultMode = X386Val.str;
  else
    X386ConfigError("Bad default mode");

  /*
   * now let's keep up with some special flags
   */
  while ((token = X386Lex(SymTab)) != STRING)
    switch (token)
      {
      case LEGEND:   vgaChipSetFlags |= C_LEGEND; break;
      case SYNCCLUT: vgaChipSetFlags |= C_SYNCCLUT; break;
      default: X386ConfigError("VGA board special flag expected");
      };

  /*
   * at last we have to scan the real mode definitions
   */
  
  while((vgaNoVideoModes < NOMODES) && (token != EOF))
    {
      if (token == STRING) 
	{
	  vgaVideoModes[vgaNoVideoModes].name = X386Val.str;
	  if (strcmp(X386Val.str,DefaultMode) == 0)
	    vgaCurrentVideoMode=vgaNoVideoModes;
	}
      else
	X386ConfigError("Mode name expected");


      if (X386Lex(NULL) == NUMBER)
	{
	  /*
	   * Get Clock Index.
	   */
	  vgaVideoModes[vgaNoVideoModes].Clock = -1;
	  for (i=0; i < vgaVideoClocks; i++) 
	    if (vgaVideoClock[i] == X386Val.num)
	      vgaVideoModes[vgaNoVideoModes].Clock = i;
	  
	  if (vgaVideoModes[vgaNoVideoModes].Clock == -1)
	    X386ConfigError("No such dot-clock available");
	  
	}
      else
	X386ConfigError("Dot clock expected");


      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].HDisplay = X386Val.num;
      else
	X386ConfigError("Horizontal display expected");
      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].HSyncStart = X386Val.num;
      else
	X386ConfigError("Horizontal sync start expected");
      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].HSyncEnd = X386Val.num;
      else
	X386ConfigError("Horizontal sync end expected");
      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].HTotal = X386Val.num;
      else
	X386ConfigError("Horizontal total expected");


      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].VDisplay = X386Val.num;
      else
	X386ConfigError("Vertical display expected");
      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].VSyncStart = X386Val.num;
      else
	X386ConfigError("Vertical sync start expected");
      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].VSyncEnd = X386Val.num;
      else
	X386ConfigError("Vertical sync end expected");
      if (X386Lex(NULL) == NUMBER)
	vgaVideoModes[vgaNoVideoModes].VTotal = X386Val.num;
      else
	X386ConfigError("Vertical total expected");

      vgaVideoModes[vgaNoVideoModes].Flags = 0;
      while (((token=X386Lex(SymTab)) != EOF) && (token != STRING))
	{
	  switch(token)
	    {
	    case INTERLACE: 
	      vgaVideoModes[vgaNoVideoModes].Flags |= V_INTERLACE;
	      break;
	    case PHSYNC: 
	      vgaVideoModes[vgaNoVideoModes].Flags |= V_PHSYNC;
	      break;
	    case NHSYNC: 
	      vgaVideoModes[vgaNoVideoModes].Flags |= V_NHSYNC;
	      break;
	    case PVSYNC: 
	      vgaVideoModes[vgaNoVideoModes].Flags |= V_PVSYNC;
	      break;
	    case NVSYNC: 
	      vgaVideoModes[vgaNoVideoModes].Flags |= V_NVSYNC;
	      break;
	    case FASTRAM: 
	      vgaVideoModes[vgaNoVideoModes].Flags |= V_FASTRAM;
	      break;
	    default:
	      X386ConfigError("Videomode special flag expected");
	      break;
	    }
	}

      vgaNoVideoModes++;
    }

  /*
   * last sanity checks
   */
  if (vgaNoVideoModes < 1)
    FatalError("At least one mode MUST be specified\n");
  if (vgaCurrentVideoMode < 0)
    FatalError("Unable to find default mode\n");

}

